
/**************************************************************************
 *                                                                        *
 * This code is developed by Adam Li.  This software is an                *
 * implementation of a part of one or more MPEG-4 Video tools as          *
 * specified in ISO/IEC 14496-2 standard.  Those intending to use this    *
 * software module in hardware or software products are advised that its  *
 * use may infringe existing patents or copyrights, and any such use      *
 * would be at such party's own risk.  The original developer of this     *
 * software module and his/her company, and subsequent editors and their  *
 * companies (including Project Mayo), will have no liability for use of  *
 * this software or modifications or derivatives thereof.                 *
 *                                                                        *
 * Project Mayo gives users of the Codec a license to this software       *
 * module or modifications thereof for use in hardware or software        *
 * products claiming conformance to the MPEG-4 Video Standard as          *
 * described in the Open DivX license.                                    *
 *                                                                        *
 * The complete Open DivX license can be found at                         *
 * http://www.projectmayo.com/opendivx/license.php .                      *
 *                                                                        *
 **************************************************************************/

/************************************************************************
 *
 *  config_dialog.cpp, Encoder Engines
 *
 *  Copyright (C) 2000  DivX Networks
 * 
 *  Adam Li
 *
 *  DivX Advance Research Center <darc@projectmayo.com>
 *
 ************************************************************************/

// This are the functions to implement the dialog for configuring the codec.

// config_dialog.cpp : implementation file
//

#include "stdafx.h"
#include "DivX.h"
#include "config_dialog.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// ConfigDialog dialog


ConfigDialog::ConfigDialog(CWnd* pParent /*=NULL*/)
	: CDialog(ConfigDialog::IDD, pParent)
{
	//{{AFX_DATA_INIT(ConfigDialog)
	m_bitrate = 0;
	m_max_quant = 0;
	m_min_quant = 0;
	m_rc_period = 0;
	m_flip = FALSE;
	//}}AFX_DATA_INIT
}


void ConfigDialog::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(ConfigDialog)
	DDX_Control(pDX, IDC_SEARCH_RANGE, m_combo_search_range);
	DDX_Control(pDX, IDC_SLIDE_BITRATE, m_slider_bitrate);
	DDX_Text(pDX, IDC_BITRATE, m_bitrate);
	DDX_Text(pDX, IDC_MAX_QUANT, m_max_quant);
	DDX_Text(pDX, IDC_MIN_QUANT, m_min_quant);
	DDX_Text(pDX, IDC_RC_PERIOD, m_rc_period);
	DDX_Check(pDX, IDC_FLIP, m_flip);
	//}}AFX_DATA_MAP
}


BEGIN_MESSAGE_MAP(ConfigDialog, CDialog)
	//{{AFX_MSG_MAP(ConfigDialog)
	ON_EN_CHANGE(IDC_BITRATE, OnChangeBitrate)
	ON_WM_HSCROLL()
	ON_CBN_SELCHANGE(IDC_SEARCH_RANGE, OnSelchangeSearchRange)
	ON_EN_CHANGE(IDC_MAX_QUANT, OnChangeMaxQuant)
	ON_EN_CHANGE(IDC_MIN_QUANT, OnChangeMinQuant)
	ON_EN_CHANGE(IDC_RC_PERIOD, OnChangeRcPeriod)
	ON_BN_CLICKED(IDC_FLIP, OnFlip)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// ConfigDialog message handlers

BOOL ConfigDialog::OnInitDialog() 
{
	CDialog::OnInitDialog();
	char sSearchRange[10];

	sprintf(sSearchRange, "%d", search_range);
	
	// extra initialization here
	m_slider_bitrate.SetRange(0, 1000, FALSE);
	m_slider_bitrate.SetPos(m_bitrate);
	m_combo_search_range.SelectString(-1, sSearchRange);

	UpdateData(FALSE);
	return TRUE;  // return TRUE unless you set the focus to a control
	              // EXCEPTION: OCX Property Pages should return FALSE
}

void ConfigDialog::OnChangeBitrate() 
{
	// When the edit box for bitrate is changed, update the slider.
	UpdateData(TRUE);
	int temp = m_bitrate;
	if (temp > 1000) temp = 1000;
	if (temp < 0) temp = 0;
	m_slider_bitrate.SetPos(temp);
	UpdateData(FALSE);
}

void ConfigDialog::OnHScroll(UINT nSBCode, UINT nPos, CScrollBar* pScrollBar) 
{
	// When the slider is changed, update the edit box.
	if ((CSliderCtrl*)pScrollBar == &m_slider_bitrate) {
		m_bitrate = m_slider_bitrate.GetPos();
		UpdateData(FALSE);
		return;
	}

	CDialog::OnHScroll(nSBCode, nPos, pScrollBar);
}

void ConfigDialog::OnSelchangeSearchRange() 
{
	int search_range_index;
	char sSearchRange[10];

	// get the user selection when the search range box is changed
	UpdateData(TRUE);
	search_range_index = m_combo_search_range.GetCurSel();
	m_combo_search_range.GetLBText(search_range_index, sSearchRange);
	m_search_range = atoi(sSearchRange);	

	return;
}

void ConfigDialog::OnChangeMaxQuant() 
{
	UpdateData(TRUE);
}

void ConfigDialog::OnChangeMinQuant() 
{
	UpdateData(TRUE);
}

void ConfigDialog::OnChangeRcPeriod() 
{
	UpdateData(TRUE);
}

void ConfigDialog::OnFlip() 
{
	UpdateData(TRUE);
}
